//
// FPGA PACMAN video hardware
//
// Version : beta1
//
// Copyright(c) 2002 Tatsuyuki Satoh , All rights reserved
//
// Important !
//
// This program is freeware for non-commercial use. 
// An author does no guarantee about this program.
// You can use this under your own risk. 
//
//

//
// 1.tilemap RAM
//  * 2 x block select RAM
//  * portA 1024 x 8 for tilemap access
//  * portB 1024 x 8 for CPU access
//
// 2.color RAM
//  * 2 x block select RAM
//  * portA 1024 x 8 for tilemap access
//  * portB 1024 x 8 for CPU access
//
// 3.sprite & work ram
//  * 2 x block select RAM
//  * portA 1024 x 8 for CPU access
//  * portB  512 x16 for sprite access
//
// 4.sprite2 ram
//  * 16 x RAM16X1 Distributed RAM
//  * 16x8 for sprite2 & CPU access
//
// 5.color lockup table & sprite line RAM
//  * block select RAM , Dual port
//  * 256 x 4 for Color lookup table
//  * 256 x 4 for sprite Line RAM
//
// 6.object pattern ROM
// * external ROM
//

module video(
//CPU interface
	A,			//CPU access address
	csvram,		//chip select video RAM  4000-4fff,c000-cfff
	csxyram,	//chip select object XY RAM 5060-506f
	n_wr,
	n_rd,
	di,
	do,
// VIDEO input
	pclk, // pixel clock
	flip, // sceen flip
// VIDEO output
	hc,	// H-coubter
	vb,	// V BLANK
	hs,	// H-SYNC
	vs,	// V-SYNC
	r,g,b,	// R,G,B 8bpp out
// pattern ROM address & data
	CA,
	CD,
	CA_as,
	CA_en,
);
// interface
input [11:0] A;
input csvram,csxyram,n_wr,n_rd;
input [7:0] di;
output [7:0] do;
input pclk,flip;
output [8:0] hc;
output vb,hs,vs;
output [2:0] r,g;
output [1:0] b;
output [12:0] CA;
input  [7:0] CD;
output CA_as,CA_en;

//
// H timming generator
//
// counter = 0-255,384-511
// hblank  = 400-495
// hsync   = 416-447
//
reg [8:0] hcnt;
reg hblank,hsync;

always @(posedge pclk)
begin

	//hblank
	if( hcnt == 399)
		hblank <= 1;
	if( hcnt == 495)
		hblank <= 0;

 	//hsync
	if( hcnt == 415)
		hsync <= 1;
	if( hcnt == 447)
		hsync <= 0;

	//H counter
	if( hcnt == 255 )
		hcnt <= 384;
	else
		hcnt <= hcnt + 1;
end

//
// V timming generator
//
// counter = 0-255,504-511
// vblank  = 240-15
// vsync   = 504-511
//
reg vdiv;
reg [8:0] vcnt;
reg vblank;
wire vsync;

//
// VSYNC == 2xHSYNC
//
always @(posedge hsync)
begin
	if( vcnt == 239)
		vblank <= 1;
	if( vcnt == 15)
		vblank <= 0;

	if( vcnt == 255 )
		vcnt <= 504;
	else
		vcnt <= vcnt + 1;
end
assign vsync = vcnt[8]; //504-511 (15KHz original timming)

//
// output timming signal
//
assign hc = hcnt;
assign vb = vblank;
assign hs = hsync;
assign vs = vsync;

//
// AB[11:0] multiplexed address bus
//
//  +---------------+----+----+-----+-----+----+----+----+-----+----+----+----+----+
//  | output        |AB11|AB10| AB9 | AB8 |AB7 |AB6 |AB5 | AB4 |AB3 |AB2 |AB1 |AB0 |
//  +---------------+----+----+-----+-----+----+----+----+-----+----+----+----+----+
//  |CPU Address    | A11| A10|  A9 |  A8 | A7 | A6 | A5 |  A4 | A3 | A2 | A1 | A0 |2H=0
//  |object Address | 1  | 1  |  1  |  1  | 1  | 1  | 1  |  1  |64H |32H |16H | 4H |2H=1,HC=400-495
//  |Left&Right tile| 0  | 4H | 64Hp| 64Hp|64Hp|64Hp| 8Hp|128Vp|64Vp|32Vp|16Vp| 8Vp|2H=1,HC=384-399,496-511
//  |Center tile    | 0  | 4H |128Vp|128Vp|32Vp|16Vp| 8Vp|128Hp|64Hp|32Hp|16Hp| 8Hp|2H=1,HC=0-255
//  +---------------+----+----+-----+-----+----+----+----+-----+----+----+----+----+
//
// parts 5S
//
// flipped address
wire [7:3] hcp = hcnt[7:3] ^ { flip,flip,flip,flip,flip };
wire [7:3] vcp = vcnt[7:3] ^ { flip,flip,flip,flip,flip };
// address bus
wire [11:0] ab_obj = { 8'b11111111,hcnt[6:4],hcnt[2]}; // object draw
wire [11:0] ab_lr  = { 1'b0,hcnt[2],hcp[6],hcp[6],hcp[6],hcp[6],hcp[3],vcp[7:3] }; // L&R tile
wire [11:0] ab_ctr = { 1'b0,hcnt[2],vcp[7:3],hcp[7:3] }; // center tile
// multiplex
wire [11:0] AB_tile  = hcnt[8] ? ab_lr    : ab_ctr; // L,R    / center
wire [11:0] AB_video = hblank  ? ab_obj   : AB_tile;// object / tile
wire [11:0] AB       = hcnt[1] ? AB_video : A[11:0];// video  / CPU

//
// CPU write enable
//
wire cpu_we = ~n_wr & (hcnt[1:0] == 2'b01); // 2H raise

//
//
// VRAM Address decoder
//
//
wire cs_code = AB[11:10] == 2'b00; // 4000-43ff , tile code RAM
wire cs_col  = AB[11:10] == 2'b01; // 4400-47ff , tile color RAM
wire cs_obj  = AB[11:10] == 2'b11; // 4c00-4fff , work / object RAM
wire code_we = csvram & cs_code & cpu_we;
wire col_we  = csvram & cs_col  & cpu_we;
wire obj_we  = csvram & cs_obj  & cpu_we;

wire [7:0] code_do; // output DATA from tile code RAM
wire [7:0] col_do;  // output DATA from color RAM
wire [7:0] obj_do;  // output DATA from work RAM + object code/color RAM
wire [7:0] DR;      // output DATA from object XY RAM
wire we_yx = csxyram & cpu_we; // WE signal to object XY RAM
wire yxram_ce = hblank | ~hcnt[1]; // chip enable to object XY RAM

//
// DBo[7:0] : shared data bus out
//
// parts 6S
//
wire [7:0] DBo = code_do | col_do | obj_do;
// data output for cpu
reg  [7:0] cpu_do;
always @(posedge pclk)
begin
//	if( hcnt[1:0] == 2'b01) // (2H raise)
	if( hcnt[1] == 1'b0) // (2H Low)
	begin
		cpu_do <= DBo;
	end
end
assign do = (csvram & ~n_rd) ? cpu_do : 8'h00;

//
// object V match & offset ,or tile V pos
//
// parts 1F,2F,3E,4D,1H,2H,4E,5L,5H,4F 
//

reg  [7:0] tile_code;  // latch of tile code / object code,flipV,flipH
reg match;             // object Y position match
reg [3:0] cv;          // V address low
reg obj_sel;           // 4H delayed HBLANK == tile / object selector

// object V+Y or tile V pos
wire [7:0] cv_add = yxram_ce ? vcnt[7:0]+DR : vcnt[7:0];
always @(posedge pclk)
begin
	if( hcnt[2:0] == 3'b011) // (4H raise)
	begin
		// latch tile code / object code+flip data from RAM
		tile_code <= DBo;
		// latch Y position , match & delayed hblank
		cv <= cv_add[3:0];
		// latch object V match flag
		match <= (cv_add[7:4] == 4'b1111);
		// latch hblank for object selector
		obj_sel <= hblank;
	end
end

//
// pixel data ROM address geneator
//
wire flipv = obj_sel ? tile_code[1] : flip; // V flip : object / tile 
wire fliph = obj_sel ? tile_code[0] : flip; // H flip : object / tile
//
assign CA[12]   = obj_sel;
assign CA[11:6] = tile_code[7:2];                       // object code / tile code[7:2]
assign CA[5]    = obj_sel ? cv[3]^flipv : tile_code[1]; // object 8V   / tile code[1]
assign CA[4]    = obj_sel ? hcnt[3]     : tile_code[0]; // object 8H   / tile code[0]
assign CA[3]    = hcnt[2] ^ fliph;                      // 4H byte select
assign CA[2:0]  = cv[2:0] ^ { flipv,flipv,flipv};       // 4V,2V,1V
assign CA_as    = hcnt[1:0] == 2'b00; // CA assert timming
assign CA_en    = hcnt[1:0] == 2'b10; // CD latch timming


//
// object LRAM access enable signal , latch color data from RAM
//
// parts 3P,2P,4B,
//
reg obj_on;					// LRAM access enable
reg hblank8;				// display enable
reg [4:0] clut_sel;			// Color Lookup Page Select from color RAM
reg fliph8;

reg hcnt8_8; /////////////

always @(posedge pclk)
begin
	if(hcnt[2:0] == 3'b111) // 1H==2H==4H==1
	begin
		// delayed HBLANK == display outout enable
		hblank8 <= hblank;
		// object line RAM enable
//		obj_on <= match | ~hcnt[8];
		obj_on <= match | ~hcnt8_8;
		// latch color RAM & H-flip for pixel generator
		clut_sel <= DBo[4:0];
		fliph8   <= fliph; 
		// 
		hcnt8_8 <= hcnt[8]; ///////////////

	end
end

//
// object line RAM address generator
//
// parts 4C,1E,2E
//
// load enable for reset tile Hpos or object Hpos
wire cntr_ld = (~obj_on | hblank8) & (hcnt[3:0] == 4'b0111);
reg [7:0] RA;
always @(posedge pclk)
begin
	if( cntr_ld )
	begin
		// H pos add + latch
		RA <= yxram_ce ? DR : 255;
	end else begin
		// H pos countup
		RA <= RA + 1;
	end
end

//
//pixel data -> 2bpp pixel selector
//
// parts 5B,5C,5H,5A
//
reg [7:0] pixel_byte;
always @(posedge pclk)
begin
	if( hcnt[1:0] == 2'b11 )
	begin
		pixel_byte <= CD;
	end else begin
		if( fliph8 )
		begin // flip , right shift
			pixel_byte[6:4] <= pixel_byte[7:5];
			pixel_byte[2:0] <= pixel_byte[3:1];
		end else begin // noflip , left shift
			pixel_byte[7:5] <= pixel_byte[6:4];
			pixel_byte[3:1] <= pixel_byte[2:0];
		end
	end
end
// 2bpp pixel data
wire [1:0] pixel = fliph8 ?
				{ pixel_byte[4],pixel_byte[0] } : // flip
				{ pixel_byte[7],pixel_byte[3] } ; // no flip


//
// Color Lookup Table ROM
//
wire clut_bank = 0;     // ColorBack , solder-selecter 0,1 or hblank8
wire [7:0] clut_d;		// data
wire [3:0] pal_code;	// palette code from pixel data
wire [3:0] pal_obj;		// palette code of object LRAM


//
// priority selector
//
// parts 3A,3B,3E
//

assign pal_code = (pal_obj==4'b0000) ? clut_d[3:0] : pal_obj;

//
// object LRAM (line ram)
//
wire [3:0] lram_di = (hblank8 & ~vblank) ? pal_code : 4'b0000;

//
// Palette Lookup Table
//

wire [3:0] pal_ad = hblank8 | vblank ? 4'h0 : pal_code;
wire [7:0] pal_do;   // data output from PaletteRAM
assign { b,g,r } = pal_do; // RGB output

//
// !!!!!!!!!!!!!!!!!!!!!!!! Memory Section !!!!!!!!!!!!!!!!!!!!!!
//
//
// tile code RAM ,4000-43ff : 8bit x 1024word
//
// parts 4K,4N
//
// Address = AB[9:0], Dout = code_do ,Din = di , wr = code_we , cs = cs_code , clk = async
//
RAMB4_S4 RAM_4N (.CLK(pclk),.ADDR(AB[9:0]),.DO(code_do[7:4]),.DI(di[7:4]),.EN(1'b1),.WE(code_we),.RST(~cs_code));
RAMB4_S4 RAM_4K (.CLK(pclk),.ADDR(AB[9:0]),.DO(code_do[3:0]),.DI(di[3:0]),.EN(1'b1),.WE(code_we),.RST(~cs_code));

//
// tile color RAM ,4400-47ff : 8bit x 1024word
//
// Address = AB[9:0], Dout = col_do ,Din = di , wr = col_we , cs = cs_code , clk = async
//
RAMB4_S4 RAM_4P (.CLK(pclk),.ADDR(AB[9:0]),.DO(col_do[7:4]),.DI(di[7:4]),.EN(1'b1),.WE(col_we),.RST(~cs_col));
RAMB4_S4 RAM_4L (.CLK(pclk),.ADDR(AB[9:0]),.DO(col_do[3:0]),.DI(di[3:0]),.EN(1'b1),.WE(col_we),.RST(~cs_col));

//
// work , object code,flip & color ram , 4c00-4fff : 8bit x 1024word
//
// parts 4R,4M
//
// Address = AB[9:0], Dout = obj_do ,Din = di , wr = obj_we , cs = cs_obj , clk = async
//
RAMB4_S4 RAM_4R (.CLK(pclk),.ADDR(AB[9:0]),.DO(obj_do[7:4]),.DI(di[7:4]),.EN(1'b1),.WE(obj_we),.RST(~cs_obj));
RAMB4_S4 RAM_4M (.CLK(pclk),.ADDR(AB[9:0]),.DO(obj_do[3:0]),.DI(di[3:0]),.EN(1'b1),.WE(obj_we),.RST(~cs_obj));

//
// object YX RAM , 5060-506f : 8bit x 16word
//
// parts 3F,3H
//
// Address = AB[3:0], Dout = obj_do ,Din = di , wr = we_yx , clk = async
//
wire [8:0] yx_ab =  { 5'b00000,AB[3:0] };
RAMB4_S8 RAM_3F_3H (.CLK(pclk),.ADDR(yx_ab),.DO(DR),.DI(~di),.EN(1'b1),.WE(we_yx),.RST(1'b0));

//
// Color Lookup Table ROM : 4bit x 256word
//
// parts 4A
//
// Addr = {clut_bank,clut_sel,pixel} , Dout = clut_d , clk = async (~pclk)
//

//
// Palette ROM            : 8bit x 16word
// parts 7F
//
// Addr = pal_ad , Dout = pal_do , oe = always on , clk = async.
//

// CLUT
wire [8:0] clut_a = { 0,clut_bank,clut_sel,pixel};

//palette
wire [8:0] pal_a9 = { 5'b10000,pal_ad };

RAMB4_S8_S8 clut_rom00 (
	.CLKA(~pclk),.ADDRA(clut_a),.DOA(clut_d),.DIA(8'h0),.WEA(1'b0),.ENA(1'b1),.RSTA(1'b0),
	.CLKB( pclk),.ADDRB(pal_a9),.DOB(pal_do),.DIB(8'h0),.WEB(1'b0),.ENB(1'b1),.RSTB(1'b0));

//
// ----- PALETTE RAM async. version -----
//
// BSRAM version make half pixel delay to RGB output.
// fill a ROM code here if you dislike that.
//
//function [7:0] plut_rom;
//input [3:0] clut;
//begin
//	case ( clut )
//	 0:plut_rom = 8'hxx;
//	 1:plut_rom = 8'hxx;
//	 2:plut_rom = 8'hxx;
//	 3:plut_rom = 8'hxx;
//	 4:plut_rom = 8'hxx;
//	 5:plut_rom = 8'hxx;
//	 6:plut_rom = 8'hxx;
//	 7:plut_rom = 8'hxx;
//	 8:plut_rom = 8'hxx;
//	 9:plut_rom = 8'hxx;
//	10:plut_rom = 8'hxx;
//	11:plut_rom = 8'hxx;
//	12:plut_rom = 8'hxx;
//	13:plut_rom = 8'hxx;
//	14:plut_rom = 8'hxx;
//	15:plut_rom = 8'hxx;
//	endcase
//end
//endfunction
//assign pal_do = plut_rom(pal_ad);
//
//RAMB4_S8 clut_rom00 (
//	.CLK(~pclk),.ADDR(clut_a),.DO(clut_d),.DI(8'h0),.WE(1'b0),.EN(1'b1),.RST(1'b0));
//
// ----- PALETTE RAM async. version -----

//
// object LRAM (line ram) : 8bit x 512word
//
// parts 1A,1B,1C,1D or 2A,2B,2B,2D , 3C or 3D
//
// Addr = RA , Dout = pal_obj , Din = lram_di , we = obj_on , oe = obj_on , clk = ~pclk
//
// WR and RD access same time and same address.
//
// With BSRAM, because Weite data are outputted in the Read data,
// separate WR & RD bus with DP-RAM and wait half cycle to write timming.
//
wire [9:0] lram_a = { 2'b00, RA };
wire [3:0] lram_do;
RAMB4_S4_S4 line_ram (
	.CLKA(pclk) ,.ADDRA(lram_a),.DOA(),.DIA(lram_di)  ,.WEA(obj_on),.ENA(1'b1),.RSTA(1'b0),
	.CLKB(~pclk),.ADDRB(lram_a),.DOB(lram_do),.DIB(4'h0),.WEB(1'b0),.ENB(1'b1),.RSTB(1'b0));
assign pal_obj = obj_on ? lram_do : 4'b0000;


endmodule
